@@ -25,10 +25,11 @@
 package Users::Groups;
 
 # enum like for verbose group array positions
-my $LOGIN  = 1;
-my $PASSWD = 2;
-my $GID    = 3;
-my $USERS  = 4;
+my $i = 0;
+my $LOGIN  = $i++;
+my $PASSWD = $i++;
+my $GID    = $i++;
+my $USERS  = $i++;
 
 # quite generic data
 $group_names = "/etc/group";
@@ -41,6 +42,7 @@ $cmd_groupmod = &Utils::File::locate_tool ("groupmod");
 $cmd_delgroup = &Utils::File::locate_tool ("delgroup");
 $cmd_addgroup = &Utils::File::locate_tool ("addgroup");
 
+$cmd_usermod  = &Utils::File::locate_tool ("usermod");
 $cmd_gpasswd  = &Utils::File::locate_tool ("gpasswd");	
 $cmd_pw       = &Utils::File::locate_tool ("pw");
 
@@ -48,57 +50,119 @@ sub del_group
 {
   my ($group) = @_;
 
+  # Make backups manually, otherwise they don't get backed up.
+  &Utils::File::do_backup ($group_names);
+
   if ($Utils::Backend::tool{"system"} eq "FreeBSD")
   {
-    $command = "$cmd_pw groupdel -n \'" . $$group[$LOGIN] . "\'";
+    @command = ($cmd_pw, "groupdel", "-n", $$group[$LOGIN]);
   }
   else
   {
-    $command  = ($cmd_delgroup) ? $cmd_delgroup : $cmd_groupdel;
-    $command .= " \'" . $$group[$LOGIN] . "\'";
+    @command  = (($cmd_delgroup) ? $cmd_delgroup : $cmd_groupdel,
+                 $$group[$LOGIN]);
   }
 
-  &Utils::File::run ($command);
+  &Utils::File::run (@command);
+}
+
+# This is only for Linux and SunOS,
+# pw groupadd manages this in FreeBSD
+sub add_user_to_group
+{
+  my ($group, $user) = @_;
+  my (@command);
+
+  if ($Utils::Backend::tool{"system"} eq "SunOS")
+  {
+    my ($groups, @arr);
+
+    $groups = &Utils::File::run_backtick ("groups $user");
+    $groups =~ s/.*://;
+    chomp ($groups);
+
+    @arr = split (/ /, $groups);
+    push @arr, $group;
+    $groups = join (',', @arr);
+    $groups =~ s/^,//;
+    $groups =~ s/,$//;
+
+    @command = ($cmd_usermod, "-G", $groups, $user);
+  }
+  else
+  {
+    @command = ($cmd_gpasswd, "-a", $user, $group);
+  }
+
+  &Utils::File::run (@command);
+}
+
+# This is only for Linux and SunOS,
+# pw groupdel manages this in FreeBSD
+sub delete_user_from_group
+{
+  my ($group, $user) = @_;
+  my (@command);
+
+  if ($Utils::Backend::tool{"system"} eq "SunOS")
+  {
+    my ($groups, @groups_arr);
+
+    $groups = &Utils::File::run_backtick ("groups $user");
+    $groups =~ s/.*://;
+    chomp ($groups);
+
+    # delete the user
+    $groups =~ s/[ \t]+$group//;
+
+    @groups_arr = split (/ /, $groups);
+    
+    @command = ($cmd_usermod, "-G", @groups_arr, $user);
+  }
+  else
+  {
+    @command = ($cmd_gpasswd, "-d", $user, $group);
+  }
+
+  &Utils::File::run (@command);
 }
 
 sub add_group
 {
   my ($group) = @_;
-  my ($u, $user, $users);
+  my ($u, $user, @users);
+
+  # Make backups manually, otherwise they don't get backed up.
+  &Utils::File::do_backup ($group_names);
 
   $u = $$group[$USERS];
 
   if ($Utils::Backend::tool{"system"} eq "FreeBSD")
   {
-    $users = join (",", sort @$u);
+    @users = sort @$u;
       
-    $command = "$cmd_pw groupadd -n \'" . $$group[$LOGIN] .
-      "\' -g \'" . $$group[$GID] .
-      "\' -M \'" . $users . "\'";
+    @command = ($cmd_pw, "groupadd", "-n", $$group[$LOGIN],
+                                     "-g", $$group[$GID],
+                                     "-M", @users);
 
-    &Utils::File::run ($command);
+    &Utils::File::run (@command);
   }
   else
   {
     if ($cmd_addgroup)
     {
-      $command = "$cmd_addgroup " .
-          "--gid \'" . $$group[$GID] . "\' " . $$group[$LOGIN];
+      @command = ($cmd_addgroup, "--gid", $$group[$GID], $$group[$LOGIN]);
     }
     else
     {
-      $command = "$cmd_groupadd -g \'" . $$group[$GID] .
-          "\' " . $$group[$LOGIN];
+      @command = ($cmd_groupadd, "-g", $$group[$GID], $$group[$LOGIN]);
     }
 
-    &Utils::File::run ($command);
+    &Utils::File::run (@command);
 
     foreach $user (sort @$u)
     {
-      $command = "$cmd_gpasswd -a \'" . $user .
-          "\' " . $$group[$LOGIN];
-
-      &Utils::File::run ($command);
+      &add_user_to_group ($$group[$LOGIN], $user);
     }
   }
 }
@@ -115,20 +179,20 @@ sub change_group
     $users_arr = $$new_group[$USERS];
     $str = join (",", sort @$users_arr);
 
-    $command = "$cmd_pw groupmod -n \'" . $$old_group[$LOGIN] .
-        "\' -g \'" . $$new_group[$GID] .
-        "\' -l \'" . $$new_group[$LOGIN] .
-        "\' -M \'" . $str . "\'";
+    @command = ($cmd_pw, "groupmod", "-n", $$old_group[$LOGIN],
+                                     "-g", $$new_group[$GID],
+                                     "-l", $$new_group[$LOGIN],
+                                     "-M", $str);
 
-    &Utils::File::run ($command);
+    &Utils::File::run (@command);
   }
   else
   {
-    $command = "$cmd_groupmod -g \'" . $$new_group[$GID] .
-        "\' -n \'" . $$new_group[$LOGIN] . "\' " .
-        "\'" . $$old_group[$LOGIN] . "\'";
+    @command = ($cmd_groupmod, "-g", $$new_group[$GID],
+                               "-n", $$new_group[$LOGIN],
+                                     $$old_group[$LOGIN]);
   
-    &Utils::File::run ($command);
+    &Utils::File::run (@command);
 
     # Let's see if the users that compose the group have changed.
     if (!Utils::Util::struct_eq ($$new_group[$USERS], $$old_group[$USERS]))
@@ -144,31 +208,56 @@ sub change_group
         {
           # users with state 2 are those that only appeared
           # in the old group configuration, so we must delete them
-          $command = "$cmd_gpasswd -d \'" . $user . "\' \'" . 
-              $$new_group[$LOGIN] . "\'";
-
-          &Utils::File::run ($command);
+          &delete_user_from_group ($$new_group [$LOGIN], $user);
         }
-        else
+        elsif ($state == 1)
         {
           # users with state 1 are those who were added
           # to the new group configuration
-          $command = "$cmd_gpasswd -a \'" . $user . "\' \'" . 
-              $$new_group[$LOGIN] . "\'";
-
-          &Utils::File::run ($command);
+          &add_user_to_group ($$new_group[$LOGIN], $user);
         }
       }
     }
   }
 }
 
+sub get_group
+{
+  my ($login) = @_;
+  my ($groups) = &get ();
+
+  foreach $group (@$groups)
+  {
+    next if ($login != $$group[$LOGIN]);
+    return $group;
+  }
+
+  return NULL;
+}
+
+sub set_group
+{
+  my ($new_group) = @_;
+  my ($groups) = &get ();
+
+  # Make backups manually, otherwise they don't get backed up.
+  &Utils::File::do_backup ($group_names);
+
+  foreach $group (@$groups)
+  {
+    if ($$new_group[$GID] == $$group[$GID])
+    {
+      &change_group ($group, $new_group);
+      return;
+    }
+  }
+}
+
 sub get
 {
   my ($ifh, @groups, $group_last_modified);
-  my (@line, $copy, @a, $counter);
+  my (@line, $copy, @a);
 
-  $counter = 1;
   $ifh = &Utils::File::open_read_from_names($group_names);
   return unless ($ifh);
 
@@ -182,13 +271,11 @@ sub get
     # FreeBSD allows comments in the group file. */
     next if &Utils::Util::ignore_line ($_);
 
-    @line = split ':', $_, -1;
-    unshift @line, $counter;
+    @line = split ':', $_, 4;
     @a = split ',', pop @line;
     push @line, [@a];
     $copy = [@line];
     push (@groups, $copy);
-    $counter++;
   }
 
   &Utils::File::close_file ($ifh);
@@ -207,49 +294,14 @@ sub get_files
 sub set
 {
   my ($config) = @_;
-  my ($old_config, %groups);
-  my (%config_hash, %old_config_hash);
 
-  if ($config)
+  return if (!$config);
+
+  # Change groups that are present in both old and new config.
+  # Groups won't be removed or added this way, for more safety.
+  foreach $group (@$config)
   {
-    # Make backup manually, otherwise they don't get backed up.
-    &Utils::File::do_backup ($group_names);
-
-    $old_config = &get ();
-
-    foreach $i (@$config) 
-    {
-      $groups{$$i[0]} |= 1;
-      $config_hash{$$i[0]} = $i;
-	  }	
-	
-    foreach $i (@$old_config)
-    {
-	    $groups{$$i[0]} |= 2;
-      $old_config_hash{$$i[0]} = $i;
-    }
-
-    # Delete all groups that only appeared in the old configuration
-    foreach $i (sort (keys (%groups)))
-    {
-      $state = $groups{$i};
-
-      if ($state == 1)
-      {
-        # Groups with state 1 have been added to the config
-        &add_group ($config_hash{$i});
-      }
-      elsif ($state == 2)
-      {
-        # Groups with state 2 have been deleted from the config
-        &del_group ($old_config_hash{$i});
-      }
-      elsif (($state == 3) &&
-             (!Utils::Util::struct_eq ($config_hash{$i}, $old_config_hash{$i})))
-      {
-        &change_group ($old_config_hash{$i}, $config_hash{$i});
-      }
-    }
+    set_group ($group);
   }
 }
 
