@@ -264,6 +264,7 @@ my $logindefs_dist_map = {
   'users_read_profiledb_fail'    => ['warn', 'Profiles read failed.'],
   'users_read_users_success'     => ['info', 'Users read successfully.'],
   'users_read_users_fail'        => ['warn', 'Users read failed.'],
+  'users_read_users_invalid'     => ['warn', 'Invalid user found while reading (missing fields).'],
   'users_read_groups_success'    => ['info', 'Groups read successfully.'],
   'users_read_groups_fail'       => ['warn', 'Groups read failed.'],
   'users_read_shells_success'    => ['info', 'Shells read successfully.'],
@@ -346,7 +347,7 @@ sub get_logindefs
 
 sub get
 {
-  my ($ifh, @users, %users_hash);
+  my ($ifh, @users, %users_hash, $fd, @passwd_status);
   my (@line, @users);
 
   # Find the passwd file.
@@ -361,6 +362,14 @@ sub get
 
     @line  = split ':', $_, -1;
     $login = $line[$LOGIN];
+
+    # skip invalid users, else they will create troubles
+    if (($line[$LOGIN] eq "") || ($line[$UID] eq ""))
+    {
+      &Utils::Report::do_report ("users_read_users_invalid");
+      next;
+    }
+
     @comment = split ',', $line[$COMMENT], 5;
 
     # we need to make sure that there are 5 elements
@@ -372,11 +381,30 @@ sub get
 
     $users_hash{$login} = [@line];
 
+    # Detect lock status of password
+    # We run 'passwd' instead of reading /etc/shadow directly
+    # to avoid leaving sensitive data in memory (hard to clear in perl)
+    $fd = &Utils::File::run_pipe_read ("passwd -S $login");
+    @passwd_status = split ' ', <$fd>;
+    &Utils::File::close_file ($fd);
+
+    if ($passwd_status[1] eq "P")
+    {
+      $users_hash{$login}[$PASSWD_STATUS] = 0;
+    }
+    elsif ($passwd_status[1] eq "NP")
+    {
+      $users_hash{$login}[$PASSWD_STATUS] = 1;
+    }
+    else # "L", means locked password
+    {
+      $users_hash{$login}[$PASSWD_STATUS] = 1 << 1;
+    }
+
     # max value for an unsigned 32 bits integer means no main group
     $users_hash{$login}[$GID] = 0xFFFFFFFF if (!$users_hash{$login}[$GID]);
 
     # TODO: read actual values
-    $users_hash{$login}[$PASSWD_STATUS] = 0;
     $users_hash{$login}[$ENC_HOME] = 0;
     $users_hash{$login}[$HOME_FLAGS] = 0;
     $users_hash{$login}[$LOCALE] = "";
@@ -466,9 +494,16 @@ sub change_user_chfn
 
 sub set_passwd
 {
-  my ($login, $password) = @_;
+  my ($login, $password, $passwd_status) = @_;
   my ($pwdpipe);
 
+  # handle empty password via passwd, as all tools don't support it
+  if ($passwd_status & 1)
+  {
+    &Utils::File::run ("passwd", "-d", $login);
+    return;
+  }
+
   if ($Utils::Backend::tool{"system"} eq "FreeBSD")
   {
     my ($command);
@@ -493,6 +528,22 @@ sub set_passwd
   }
 }
 
+# Enable/disable password, only call if value has changed
+sub set_lock
+{
+  my ($login, $passwd_status) = @_;
+  my ($pwdpipe);
+
+  if ($passwd_status & (1 << 1))
+  {
+    &Utils::File::run ("passwd", "-l", $login);
+  }
+  else
+  {
+    &Utils::File::run ("passwd", "-u", $login);
+  }
+}
+
 # This function allows empty values to be passed, in which cas
 # the platform's tools will choose the default.
 sub add_user
@@ -594,6 +645,10 @@ sub add_user
     }
   }
 
+  &change_user_chfn ($$user[$LOGIN], undef, $$user[$COMMENT]);
+  &set_passwd ($$user[$LOGIN], $$user[$PASSWD], $$user[$PASSWD_STATUS]);
+  &set_lock ($$user[$LOGIN], $$user[$PASSWD_STATUS]);
+
   $chown_home = $$user[$HOME_FLAGS] & (1 << 1);
 
   # update user to get values that were filled
@@ -606,9 +661,6 @@ sub add_user
     &Utils::File::run (@command);
   }
 
-  &change_user_chfn ($$user[$LOGIN], undef, $$user[$COMMENT]);
-  &set_passwd ($$user[$LOGIN], $$user[$PASSWD]);
-
   # Return the new user with default values filled.
   # Returns NULL if user doesn't exist, which means failure.
   return $user;
@@ -642,7 +694,13 @@ sub change_user
   }
 
   &change_user_chfn ($$new_user[$LOGIN], $$old_user[$COMMENT], $$new_user[$COMMENT]);
-  &set_passwd ($$new_user[$LOGIN], $$new_user[$PASSWD]);
+  &set_passwd ($$new_user[$LOGIN], $$new_user[$PASSWD], $$user[$PASSWD_STATUS]);
+
+  # Only change lock status if status has changed
+  if (($$new_user[$PASSWD_STATUS] & (1 << 1)) != ($$user[$PASSWD_STATUS] & (1 << 1)))
+  {
+    &set_lock ($$user[$LOGIN], $$new_user[$PASSWD_STATUS]);
+  }
 
   # Erase password string to avoid it from staying in memory
   $$new_user[$PASSWD] = '0' x length ($$new_user[$PASSWD]);
