@@ -15,7 +15,8 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
  *
- * Authors: Carlos Garnacho Parro  <carlosg@gnome.org>
+ * Authors: Carlos Garnacho Parro  <carlosg@gnome.org>,
+ *          Milan Bouchet-Valat <nalimilan@club.fr>.
  */
 
 #include <unistd.h>
@@ -30,10 +31,7 @@
 #include "config.h"
 #include "dispatcher.h"
 
-#ifdef HAVE_POLKIT
 #include <polkit/polkit.h>
-#include <polkit-dbus/polkit-dbus.h>
-#endif
 
 #ifdef HAVE_GIO
 #include "file-monitor.h"
@@ -42,8 +40,8 @@
 #define DBUS_ADDRESS_ENVVAR "DBUS_SESSION_BUS_ADDRESS"
 #define DBUS_INTERFACE_STB "org.freedesktop.SystemToolsBackends"
 #define DBUS_INTERFACE_STB_PLATFORM "org.freedesktop.SystemToolsBackends.Platform"
-#define DBUS_PATH_USER_CONFIG "/org/freedesktop/SystemToolsBackends/UserConfig"
-#define DBUS_PATH_SELF_CONFIG "/org/freedesktop/SystemToolsBackends/SelfConfig"
+#define DBUS_INTERFACE_STB_AUTH "org.freedesktop.SystemToolsBackends.Authentication"
+#define DBUS_PATH_SELF_CONFIG "/org/freedesktop/SystemToolsBackends/SelfConfig2"
 
 #define STB_DISPATCHER_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), STB_TYPE_DISPATCHER, StbDispatcherPrivate))
 
@@ -66,9 +64,7 @@ struct StbDispatcherPrivate
   DBusConnection *connection;
   gchar *platform;
 
-#ifdef HAVE_POLKIT
-  PolKitContext *polkit_context;
-#endif
+  PolkitAuthority *polkit_authority;
 
 #ifdef HAVE_GIO
   StbFileMonitor *file_monitor;
@@ -312,71 +308,149 @@ get_destination (DBusMessage *message)
   return destination;
 }
 
+static void
+return_error (StbDispatcher *dispatcher,
+	      DBusMessage   *message,
+	      const gchar   *error_name,
+	      const gchar   *error_message)
+{
+  DBusMessage *reply;
+  StbDispatcherPrivate *priv;
+
+  priv = STB_DISPATCHER_GET_PRIVATE (dispatcher);
+
+  DEBUG (dispatcher, "sending error %s from: %s", error_name, dbus_message_get_path (message));
+
+  reply = dbus_message_new_error (message, error_name, error_message);
+  dbus_connection_send (priv->connection, reply, NULL);
+  dbus_message_unref (reply);
+}
+
+static gboolean
+check_polkit_auth (StbDispatcher *dispatcher,
+                   DBusMessage   *message,
+                   GError       **ret_error)
+{
+  StbDispatcherPrivate *priv;
+  PolkitSubject *subject;
+  PolkitAuthorizationResult *result;
+  GError *error;
+  gchar **path;
+  gchar *action_id;
+  gboolean retval;
+
+  priv = STB_DISPATCHER_GET_PRIVATE (dispatcher);
+
+  if (dbus_message_get_path_decomposed (message, &path)
+      && path[0] && path[1] && path[2] && path[3]
+      && strcmp (path[3], "SelfConfig2") == 0)
+    action_id = g_strdup_printf ("org.freedesktop.systemtoolsbackends.self.set");
+  else
+    action_id = g_strdup_printf ("org.freedesktop.systemtoolsbackends.set");
+
+  subject = polkit_system_bus_name_new (dbus_message_get_sender (message));
+  result = polkit_authority_check_authorization_sync (priv->polkit_authority, subject, action_id, NULL,
+                                                      POLKIT_CHECK_AUTHORIZATION_FLAGS_ALLOW_USER_INTERACTION,
+                                                      NULL, &error);
+  if (result)
+    {
+      retval = polkit_authorization_result_get_is_authorized (result);
+      g_object_unref (result);
+    }
+  else
+    {
+      retval = FALSE;
+
+      if (error->code == POLKIT_ERROR_CANCELLED)
+          /* Clients need to hanle this case separately to avoid showing an error dialog */
+          return_error (dispatcher, message,
+                        "org.freedesktop.SystemToolsBackends.AuthenticationCancelled",
+                        "Authentication cancelled by user");
+      else
+          return_error (dispatcher, message,
+                        "org.freedesktop.SystemToolsBackends.AuthenticationFailed",
+                        error->message);
+
+        g_warning ("Error checking PolicyKit authorization: %s", error->message);
+
+        if (!ret_error)
+          g_error_free (error);
+        else
+          *ret_error = error;
+    }
+
+  DEBUG (dispatcher,
+	 (retval) ? "subject is allowed to do action '%s'" : "subject can't do action '%s'",
+	 action_id);
+
+  g_free (action_id);
+  g_object_unref (subject);
+
+  if (path)
+    g_strfreev (path);
+
+  return retval;
+}
+
 static gboolean
 can_caller_do_action (StbDispatcher *dispatcher,
-		      DBusMessage   *message,
-		      const gchar   *name)
+		      DBusMessage   *message)
 {
-#ifdef HAVE_POLKIT
-  StbDispatcherPrivate *priv;
-  PolKitAction *action;
-  PolKitCaller *caller;
-  DBusError error;
-  PolKitResult result;
-  const gchar *member;
-  gchar *action_id;
   gboolean retval;
 
   /* Allow getting information */
   if (dbus_message_has_member (message, "get"))
     return TRUE;
 
-  /* Do not allow anything besides "set" past this point */
-  if (!dbus_message_has_member (message, "set"))
+  /* Do not allow anything besides "set", "add" or "del" past this point */
+  if (!(dbus_message_has_member (message, "set")
+       || dbus_message_has_member (message, "add")
+       || dbus_message_has_member (message, "del")))
     return FALSE;
 
-  priv = STB_DISPATCHER_GET_PRIVATE (dispatcher);
-  member = dbus_message_get_member (message);
-
-  g_return_val_if_fail (member != NULL, FALSE);
-
-  action = polkit_action_new ();
-
-  if (name)
-    action_id = g_strdup_printf ("org.freedesktop.systemtoolsbackends.%s.%s", name, member);
-  else
-    action_id = g_strdup_printf ("org.freedesktop.systemtoolsbackends.%s", member);
-
-  polkit_action_set_action_id (action, action_id);
-
-  dbus_error_init (&error);
-  caller = polkit_caller_new_from_dbus_name (priv->connection, dbus_message_get_sender (message), &error);
-
-  if (dbus_error_is_set (&error))
-    {
-      g_critical ("%s", error.message);
-      dbus_error_free (&error);
-
-      return FALSE;
-    }
-
-  result = polkit_context_can_caller_do_action (priv->polkit_context, action, caller);
-
-  polkit_caller_unref (caller);
-  polkit_action_unref (action);
-
-  retval = (result == POLKIT_RESULT_YES);
-
-  DEBUG (dispatcher,
-	 (retval) ? "caller is allowed to do action '%s'" : "caller can't do action '%s'",
-	 action_id);
-
-  g_free (action_id);
+  /* Replies with an error if needed, in which case retval is FALSE */
+  retval = check_polkit_auth (dispatcher, message, NULL);
 
   return retval;
-#else
-  return TRUE;
-#endif /* HAVE_POLKIT */
+}
+
+/*
+ * Handles all messages with DBUS_INTERFACE_STB_AUTH, which check authorizations
+ * for all modules.
+ */
+static void
+dispatch_auth_message (StbDispatcher *dispatcher,
+                       DBusMessage   *message)
+{
+  StbDispatcherPrivate *priv;
+  GError *error = NULL;
+
+  priv = dispatcher->_priv;
+
+  if (!dbus_message_has_interface (message, DBUS_INTERFACE_STB_AUTH))
+    return;
+
+  if (dbus_message_has_member (message, "authenticate"))
+    {
+      DBusMessage *reply;
+      DBusMessageIter iter;
+      gboolean authorized;
+
+      /* Replies with an error if needed */
+      authorized = check_polkit_auth (dispatcher, message, &error);
+
+      if (!error)
+        {
+           /* Create a reply with the authorization result */
+           reply = dbus_message_new_method_return (message);
+
+           dbus_message_iter_init_append (reply, &iter);
+           dbus_message_iter_append_basic (&iter, DBUS_TYPE_BOOLEAN, &authorized);
+
+           dbus_connection_send (priv->connection, reply, NULL);
+           dbus_message_unref (reply);
+        }
+    }
 }
 
 static void
@@ -423,24 +497,6 @@ dispatch_stb_message (StbDispatcher *dispatcher,
   dbus_message_unref (copy);
 }
 
-static void
-return_error (StbDispatcher *dispatcher,
-	      DBusMessage   *message,
-	      const gchar   *error_name)
-{
-  DBusMessage *reply;
-  StbDispatcherPrivate *priv;
-
-  priv = STB_DISPATCHER_GET_PRIVATE (dispatcher);
-
-  DEBUG (dispatcher, "sending error %s from: %s", error_name, dbus_message_get_path (message));
-
-  reply = dbus_message_new_error (message, error_name,
-				  "No permissions to perform the task.");
-  dbus_connection_send (priv->connection, reply, NULL);
-  dbus_message_unref (reply);
-}
-
 static void
 dispatch_platform_message (StbDispatcher *dispatcher,
 			   DBusMessage   *message)
@@ -474,78 +530,32 @@ dispatch_platform_message (StbDispatcher *dispatcher,
 }
 
 static void
-dispatch_user_config (StbDispatcher *dispatcher,
+dispatch_self_config (StbDispatcher *dispatcher,
 		      DBusMessage   *message)
 {
   StbDispatcherPrivate *priv;
-  DBusMessageIter iter;
-  DBusMessage *user_message = NULL;
   const gchar *sender;
-  gulong uid;
+  gulong uid, message_uid;
 
   priv = dispatcher->_priv;
   sender = dbus_message_get_sender (message);
-  uid = (uid_t) dbus_bus_get_unix_user (priv->connection, sender, NULL);
+  uid = dbus_bus_get_unix_user (priv->connection, sender, NULL);
 
-  g_return_if_fail (uid != -1);
+  /* Absolutely avoid passing UID 0 */
+  g_return_if_fail (uid > 0);
 
-  if (dbus_message_has_member (message, "get"))
+  if (dbus_message_get_args (message, NULL,
+                             DBUS_TYPE_UINT32, &message_uid,
+                             DBUS_TYPE_INVALID)
+                             && message_uid == uid)
     {
-      /* compose the call to UserConfig with the uid of the caller */
-      user_message = dbus_message_new_method_call (DBUS_INTERFACE_STB ".UserConfig",
-						   DBUS_PATH_USER_CONFIG,
-						   DBUS_INTERFACE_STB,
-						   "get");
-      dbus_message_iter_init_append (user_message, &iter);
-      dbus_message_iter_append_basic (&iter, DBUS_TYPE_UINT32, &uid);
-    }
-  else if (dbus_message_has_member (message, "set"))
-    {
-      DBusMessageIter array_iter;
-      const gchar *passwd;
-      gchar **gecos;
-      gint gecos_elements, i;
-      uid_t message_uid;
-
-      if (dbus_message_get_args (message, NULL,
-				 DBUS_TYPE_UINT32, &message_uid,
-				 DBUS_TYPE_STRING, &passwd,
-				 DBUS_TYPE_ARRAY, DBUS_TYPE_STRING, &gecos, &gecos_elements,
-				 DBUS_TYPE_INVALID))
-	{
-	  user_message = dbus_message_new_method_call (DBUS_INTERFACE_STB ".UserConfig",
-						       DBUS_PATH_USER_CONFIG,
-						       DBUS_INTERFACE_STB,
-						       "set");
-
-	  dbus_message_iter_init_append (user_message, &iter);
-	  dbus_message_iter_append_basic (&iter, DBUS_TYPE_UINT32, &uid);
-	  dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &passwd);
-
-	  /* append gecos array */
-	  dbus_message_iter_open_container (&iter,
-					    DBUS_TYPE_ARRAY,
-					    DBUS_TYPE_STRING_AS_STRING,
-					    &array_iter);
-
-	  for (i = 0; gecos[i]; i++)
-	    dbus_message_iter_append_basic (&array_iter, DBUS_TYPE_STRING, &gecos[i]);
-
-	  dbus_message_iter_close_container (&iter, &array_iter);
-	  dbus_free_string_array (gecos);
-	}
+      dbus_message_set_sender (message, sender);
+      dispatch_stb_message (dispatcher, message, 0);
     }
   else
-    g_warning ("unsupported method on SelfConfig");
-
-  if (user_message)
-    {
-      dbus_message_set_sender (user_message, sender);
-      dispatch_stb_message (dispatcher, user_message, dbus_message_get_serial (message));
-      dbus_message_unref (user_message);
-    }
-  else
-    return_error (dispatcher, message, DBUS_ERROR_UNKNOWN_METHOD);
+    return_error (dispatcher, message,
+                  "org.freedesktop.SystemToolsBackends.InvalidMessage",
+                  "Malformed message was sent");
 }
 
 static DBusHandlerResult
@@ -567,19 +577,19 @@ dispatcher_filter_func (DBusConnection *connection,
     dispatch_stb_message (dispatcher, message, 0);
   else if (dbus_message_has_interface (message, DBUS_INTERFACE_STB_PLATFORM))
     dispatch_platform_message (dispatcher, message);
+  else if (dbus_message_has_interface (message, DBUS_INTERFACE_STB_AUTH))
+    dispatch_auth_message (dispatcher, message);
   else if (dbus_message_has_path (message, DBUS_PATH_SELF_CONFIG))
     {
-      if (can_caller_do_action (dispatcher, message, "self"))
-	dispatch_user_config (dispatcher, message);
-      else
-	return_error (dispatcher, message, DBUS_ERROR_ACCESS_DENIED);
+      /* Error is returned from can_caller_do_action() in case of failure */
+      if (can_caller_do_action (dispatcher, message))
+	dispatch_self_config (dispatcher, message);
     }
   else if (dbus_message_has_interface (message, DBUS_INTERFACE_STB))
     {
-      if (can_caller_do_action (dispatcher, message, NULL))
+      /* Error is returned from can_caller_do_action() in case of failure */
+      if (can_caller_do_action (dispatcher, message))
 	dispatch_stb_message (dispatcher, message, 0);
-      else
-	return_error (dispatcher, message, DBUS_ERROR_ACCESS_DENIED);
     }
 
   return DBUS_HANDLER_RESULT_HANDLED;
@@ -615,48 +625,6 @@ setup_connection (StbDispatcher *dispatcher)
   dbus_connection_setup_with_g_main (priv->connection, NULL);
 }
 
-#ifdef HAVE_POLKIT
-static gboolean
-stb_polkit_io_watch_func (GIOChannel   *channel,
-			  GIOCondition  condition,
-			  gpointer      user_data)
-{
-  int fd;
-  PolKitContext *pk_context;
-
-  pk_context = (PolKitContext *) user_data;
-  fd = g_io_channel_unix_get_fd (channel);
-  polkit_context_io_func (pk_context, fd);
-
-  return TRUE;
-}
-
-static int
-stb_polkit_io_add_watch (PolKitContext *context,
-			 int            fd)
-{
-  guint watch_id = 0;
-  GIOChannel *channel;
-
-  channel = g_io_channel_unix_new (fd);
-
-  if (!channel)
-    return 0;
-
-  watch_id = g_io_add_watch (channel, G_IO_IN, stb_polkit_io_watch_func, context);
-  g_io_channel_unref (channel);
-
-  return watch_id;
-}
-
-static void
-stb_polkit_io_remove_watch (PolKitContext *context,
-			    int            watch_id)
-{
-  g_source_remove (watch_id);
-}
-#endif
-
 static void
 stb_dispatcher_init (StbDispatcher *dispatcher)
 {
@@ -670,14 +638,7 @@ stb_dispatcher_init (StbDispatcher *dispatcher)
   /* we're screwed if we don't have this */
   g_assert (priv->connection != NULL);
 
-#ifdef HAVE_POLKIT
-  priv->polkit_context = polkit_context_new ();
-  polkit_context_set_io_watch_functions (priv->polkit_context,
-					 stb_polkit_io_add_watch,
-					 stb_polkit_io_remove_watch);
-
-  polkit_context_init (priv->polkit_context, NULL);
-#endif
+  priv->polkit_authority = polkit_authority_get ();
 
 #ifdef HAVE_GIO
   priv->file_monitor = stb_file_monitor_new ();
@@ -730,9 +691,7 @@ stb_dispatcher_finalize (GObject *object)
 
   dbus_connection_unref (priv->connection);
 
-#ifdef HAVE_POLKIT
-  polkit_context_unref  (priv->polkit_context);
-#endif
+  g_object_unref (priv->polkit_authority);
 
 #ifdef HAVE_GIO
   g_object_unref (priv->file_monitor);
